'use strict';

// Declare app level module which depends on views, and components
angular.module('algo', [
  'ngRoute',
  'ui.bootstrap',
  'colorpicker.module',
  'algo.scheduler'
])

.config(['$routeProvider', function($routeProvider) {
  $routeProvider.otherwise({redirectTo: '/calendar'});
}])

// Shim to replicate Angular 1.3's ng-model-options="{ updateOn: 'blur' }"
.directive('ngModelOnblur', function() {
  return {
    priority: 1,
    restrict: 'A',
    require: 'ngModel',
    link: function(scope, elm, attr, ngModelCtrl) {
      if (attr.type === 'radio' || attr.type === 'checkbox') return;

      elm.off('input keydown change');
      elm.on('blur', function() {
        scope.$apply(function() {
          ngModelCtrl.$setViewValue(elm.val());
        });
      });
    }
  };
})

// From http://stackoverflow.com/questions/11873570/angularjs-for-loop-with-numbers-ranges
.filter('makeRange', function() {
  return function(input) {
    var lowBound, highBound;
    switch (input.length) {
    case 1:
      lowBound = 0;
      highBound = parseInt(input[0]) - 1;
      break;
    case 2:
      lowBound = parseInt(input[0]);
      highBound = parseInt(input[1]);
      break;
    default:
      return input;
    }
    var result = [];
    for (var i = lowBound; i <= highBound; i++)
      result.push(i);
    return result;
  };
})

/**
 * A feature to allow different views to share the same
 * data. Useful right now only due to client-side, server
 * -side separation
 */
.factory('shared', ['$http', 'moment', function sharedFactory($http, moment) {
  var sharedData;

  /**
   * Turn an array of objects into a map with the object's "key" being the
   * map's key.
   * @param   array     array to convert
   * @param   key       property to use as the new key in the hash
   * @param   value     if specified, property to use as the value
   */
  var arrayToMap = function(array, key, value) {
    var map = {};
    for (var i = 0; i < array.length; i++) {
      var elem = array[i];
      var val = value ? elem[value] : elem;
      map[elem[key]] = val;
    }
    return map;
  };

  /**
   * Turn a map into an array, with each value of a key-value pair being added
   * @param   map       map to convert
   * @param   key       if specified, the field to assign the map's keys to
   * @param   value     if specified, the field to assign the map's values to
   */
  var mapToArray = function(map, key, value) {
    var array = [];
    $.each(map, function(k, v) {
      var obj = (key || value) ? {} : v;
      if (key) {
        obj[key] = k;
      }
      if (value) {
        obj[value] = v;
      }
      array.push(obj);
    });

    return array;
  };

  /**
   * The server uses a slightly different JSON format than the client.
   * This converts client-side scheduler.json to the server-side scheduler.json
   */
  var clientToServer = function(data) {
    var serverData = angular.copy(data);

    // 1. Convert the schedules
    $.each(data.schedules, function(id, schedule) {
      var schedule = data.schedules[id];

      // Remove the ID
      delete schedule.id;

      for (var j = 0; j < schedule.events.length; j++) {
        var event = schedule.events[j];
        event.zone = event.zone.toString();
      }
    });

    // 2. Convert the calendars
    $.each(data.calendar, function(day, scheduleId) {
      if (data.schedules[scheduleId] != null) {
        data.calendar[day] = data.schedules[scheduleId].name;
      } else {
        delete data.calendar[day];
      }
    });
    serverData.calendar = mapToArray(data.calendar, 'date', 'schedule');
    serverData.schedules = mapToArray(data.schedules);
    return serverData;
  };

  var serverToClient = function(serverData) {
    var data = angular.copy(serverData);

    // 1. Convert the schedules
    var idForName = {};
    for (var i = 0; i < data.schedules.length; i++) {
      var schedule = data.schedules[i];

      // Assign unique, uneditable IDs
      schedule.id = i + 1;

      idForName[schedule.name] = schedule.id;

      // Convert zones to numbers from strings
      for (var j = 0; j < schedule.events.length; j++) {
        var event = schedule.events[j];
        event.zone = parseInt(event.zone);
      }
    }
    data.schedules = arrayToMap(data.schedules, 'id');

    // 2. Change calendar so that "calendar" is an array keyed
    // by its "date" field and has the ID rather than the schedule name for the value
    for (var i = 0; i < data.calendar.length; i++) {
      var day = data.calendar[i];
      day.scheduleId = idForName[day.schedule];
    }
    data.calendar = arrayToMap(data.calendar, 'date', 'scheduleId');
    return data;
  };

  /**
   * Persist the data onto the server.
   * @note this can be called as persist().success(...).error(...)
   *       since the $http object is returned.
   */
  var persist = function(invalidCallback) {
    var data = angular.copy(sharedData); // Avoid dirtying

    // Edit sharedData so that the event times are in standard format ('10:05 am/pm')
    // and return false if it doesn't work

    var invalidMessages = [];

    var scheduleNames = {};

    $.each(data.schedules, function(id, schedule) {
      if (!schedule.name) {
        if (scheduleNames.length > 1)
          invalidMessages.push('Schedule name cannot remain blank; please give a unique name.');
        else
          invalidMessages.push('Schedule names cannot remain blank; please give them unique names.');
      } else if (schedule.name.length > 32) {
          invalidMessages.push('Schedule name cannot be longer than 32 characters.');
      } else if (schedule.name in scheduleNames) {
        invalidMessages.push('There are multiple schedules with the name "' + schedule.name + '"; please give them unique names.');
      }
      scheduleNames[schedule.name] = true;

      if (!schedule.events) {
        schedule.events = [];
      }

      $.each(schedule.events, function(i, event) {
        var momentTime = makeMoment(event.time);
        if (momentTime.isValid()) {
          // Standardize formats
          event.time = momentTime.format('HH:mm:ss');
        } else {
          invalidMessages.push('The time "' + event.time + '" in your schedule ' + schedule.name + ' is not a valid time.');
        }
      });
    });

    if (invalidMessages.length != 0) {
      if (invalidCallback) {
        invalidCallback(invalidMessages);
      }
      var ret = {
        // Pretend to be a promise-deliverer
        success: function() { return ret; },
        error: function() { return ret; }
      };
      return ret;
    }
    var serverData = clientToServer(data);

    return $http.post('sched-save.lua', serverData);
  };

  var fetch = function(successCallback, errorCallback) {
    if (sharedData) {
      successCallback(sharedData);
      return;
    }

    $http.get('sched-load.lua').
      success(function success(serverData, status, headers, config) {
        var data = serverToClient(serverData);
        sharedData = data;
        successCallback(sharedData);
      }).
      error(function error(data, status, headers, config) {
        if (errorCallback) {
          errorCallback(data, status, headers, config);
        }
      });
  };

  /**
   * Turn a time string into a moment.js string
   */
  var makeMoment = function(time) {
    var formats = ['H:m:s', 'h:m:s a'];
    return moment(time, formats);
  };

  var activateMenu = function(index) {
    var secondMenuUl = document.getElementsByClassName("tabmenu l2");
    var secondMenuElements = secondMenuUl[0].getElementsByTagName("li");

    for (var i = 0; i < secondMenuElements.length; i++) {
      secondMenuElements[i].classList.remove('active');
    }

    secondMenuElements[index].classList.add('active');
  };

  return {
    persist: persist,
    fetch: fetch,
    makeMoment: makeMoment,
    activateMenu: activateMenu
  };
}])

// Shim for Angular v2:

// moment.js functions
.factory('moment', function() {
  return moment; // Global variable from moment.js
})

.filter('parseDate', function() {
  return function(dateString, format) {
    return moment(dateString, format);
  };
})

.filter('dateFormat', function() {
  return function(date, format) {
    return date.format(format);
  };
});
